"""Exception classes for beka"""

class SocketClosedError(Exception):
    """Exception raised when socket closed error occurs"""

class IdleError(Exception):
    """Exception raised when idle error occurs"""
